import krister.Ess.*;
import oscP5.*;
import netP5.*;
import controlP5.*;
ControlP5 controlP5;
OscP5 oscP5;
NetAddress myRemoteLocation;

AudioInput in;
int bufferSize;
FFT analysis;
BeatDetect frame;

boolean equalize = true;
boolean limit = true;
boolean useSmooth = true;
float dampAmt = 1.0;
boolean useAvg = false;
int numAvg = 64;
float upMult = 0.95;

int kp = 0;
int sp = 0;
int hp = 0;
int op = 0;

color c1 = #FF0000;
color c2 = #FFFFFF;
color c3 = #666666;
color c4 = #333333;

int ball_x = 200;
int ball_x_offset = 80;
int ball_y = 290;

float adjust_value = 100.0;

void setup(){
  smooth();

  controlP5 = new ControlP5(this);

  controlP5.setColorActive(c1);
  controlP5.setColorLabel(c2);
  controlP5.setColorForeground(c2);
  controlP5.setColorBackground(c4);

  controlP5.addToggle("EQ",true,10,240,10,10);
  //controlP5.addToggle("Limits",true,10,270,10,10);
  controlP5.addToggle("Average",false,10,270,10,10);
  controlP5.addToggle("Mode",false,10,300,10,10);

  Slider s = controlP5.addSlider("Damping",0.0,1.0,1.0,10,330,420,10);
  s = controlP5.addSlider("Adjust",0.0,200.0,100.0,10,360,420,10);

  size(512,384);
  strokeCap(SQUARE);
  frameRate(60);

  Ess.start(this);
  Ess.masterVolume(1);
  bufferSize = 1024;
  in = new AudioInput(bufferSize);
  analysis = new FFT(bufferSize);
  analysis.smooth = useSmooth;
  analysis.equalizer(equalize);
  analysis.limits();
  analysis.damp(dampAmt);
  analysis.averages(numAvg);
  analysis.updateMultiplier = upMult;
  frame = new BeatDetect(in.size, in.sampleRate);
  frame.detectMode("FREQ_ENERGY");
  in.start();

  /* start oscP5, listening for incoming messages at port 8002 */
  oscP5 = new OscP5(this,8002);
  myRemoteLocation = new NetAddress("127.0.0.1",8002);
  //myRemoteLocation = new NetAddress("192.168.178.20",8002);
}


//******************************************************************************
//beatdetect31..

void draw()
{
  background(0);
  OscMessage myMessage = new OscMessage("/md8key/beat");

  if ( frame.isOnset() ) 
  {
    drawBall(ball_x+ball_x_offset*3, ball_y);
    op = 3;
    //data[0]=1;
    int data = 1;
    myMessage.add(data);
  }
  else if ( op > 0 )
  {
    drawBall(ball_x+ball_x_offset*3, ball_y);
    op--;
    //data[0]=0;
    int data = 0;
    myMessage.add(data);
  }

  if ( frame.isKick() )
  {
    drawBall(ball_x, ball_y);
    kp = 3;
    //data[1]=1;
    int data = 1;
    myMessage.add(data);
  }
  else if ( kp > 0 )
  {
    drawBall(ball_x, ball_y);
    kp--;
    //data[1]=0;
    int data = 0;
    myMessage.add(data);
  }

  if ( frame.isSnare() )
  {
    drawBall(ball_x+ball_x_offset, ball_y);
    sp = 3;
    //data[2]=1;
    int data = 1;
    myMessage.add(data);
  }
  else if ( sp > 0 )
  {
    drawBall(ball_x+ball_x_offset, ball_y);
    sp--;
    //data[2]=0;
    int data = 0;
    myMessage.add(data);
  }

  if ( frame.isHat() )
  {
    drawBall(ball_x+ball_x_offset*2, ball_y);
    hp = 3;
    //data[3]=1;
    int data = 1;
    myMessage.add(data);
  }
  else if ( hp > 0 )
  {
    drawBall(ball_x+ball_x_offset*2, ball_y);
    hp--;
    //data[3]=0;
    int data = 0;
    myMessage.add(data);
  }

  oscP5.send(myMessage, myRemoteLocation);
  myMessage.clear();

  // send OSC
  if (useAvg)
  {
    OscMessage myMessage2 = new OscMessage("/md8key/Average");
    for (int i = 0; i < analysis.averages.length; i++)
    {
      float data = analysis.averages[i]*adjust_value;
      myMessage2.add(data);
    }
    oscP5.send(myMessage2, myRemoteLocation);
    myMessage2.clear();
  }
  else
  {
    OscMessage myMessage3 = new OscMessage("/md8key/Spectrum");
    for (int i = 0; i < analysis.spectrum.length; i=i+4)
    {
      float data = analysis.spectrum[i]*adjust_value;
      myMessage3.add(data);
      //println (data);
    }
    oscP5.send(myMessage3, myRemoteLocation);
    myMessage3.clear();
  }


  // draw spectrum
  stroke(255);
  if (useAvg)
  {
    strokeWeight(4);
    for (int i = 0; i < analysis.averages.length; i++)
    {
      line(i*8, height*-1, i*8, analysis.averages[i]*adjust_value);
    }
  }
  else
  {
    strokeWeight(1);      
    for (int i = 0; i < analysis.spectrum.length; i=i+2)
    {
      line(i, height*-1, i, analysis.spectrum[i]*adjust_value);
    }
  }

  // draw lines for eq
  stroke(0, 0, 255);
  strokeWeight(1);
  line(0, 10, width, 10);
  stroke(0, 255, 0);
  line(0, 50, width, 50);
  stroke(255, 0, 0);
  line(0, 100, width, 100);

  // draw onset graph
  frame.drawGraph();
}

void drawBall(int x, int y)
{
  noStroke();
  fill(255);
  ellipse(x, y, 40, 40);
}

void keyPressed()
{
  if ( key == 'a' )
  {
    useAvg = (useAvg ? false : true);
  }
  if ( key == 'e' ) 
  { 
    equalize = (equalize ? false : true);
    analysis.equalizer(equalize);
  }
  if ( key == 'l' )
  {
    if (limit)
    {
      analysis.noLimits();
      limit = false;
    }
    else
    {
      analysis.limits();
      limit = true;
    }
  }
  if ( key == 'f' )
  {
    dampAmt += 0.1;
    if ( dampAmt > 1 ) dampAmt = 1;
    println("Damp amount changed to " + dampAmt + ".");
    analysis.damp(dampAmt);
    //frame.spect.damp(dampAmt);
  }
  if ( key == 'd' )
  {
    dampAmt -= 0.1;
    if ( dampAmt < 0 ) dampAmt = 0;
    println("Damp amount changed to " + dampAmt + ".");
    analysis.damp(dampAmt);
    //frame.spect.damp(dampAmt);
  }
  if ( key == '1' )
  {
    frame.detectMode("SOUND_ENERGY");
  }
  if ( key == '2' )
  {
    frame.detectMode("FREQ_ENERGY");
  }

}

public void audioInputData(AudioInput theInput) {
  analysis.getSpectrum(in);
  frame.detect(in);
}

public void stop()
{
  Ess.stop();
  super.stop();
}

//##############################################

class BeatDetect
{
  private int algorithm;
  private boolean isOnset;
  private FFT spect;
  private int bins;
  private float sampleRate;
  private int currFrame;
  private int valCnt;
  private float[] valGraph;

  // vars for sEnergy
  private float[] eBuffer;
  private float[] dBuffer;

  // vars for fEnergy
  private float[][] feBuffer;
  private float[][] fdBuffer;
  private boolean[] fIsOnset;
  private float[] varGraph;
  private int varCnt;
  private int feNum;

  BeatDetect()
  {
    init(1024, 44100);
  }

  BeatDetect(int s, float r)
  {
    init(s, r);
  }

  private void init(int s, float r)
  {
    bins = s;
    sampleRate = r;
    algorithm = 1;
    isOnset = false;
    currFrame = frameCount;
    spect = new FFT(bins*2);
    spect.damp(1.0);
    valCnt = varCnt = 0;
    valGraph = new float[width];
    varGraph = new float[width];
    // sEnergy vars: energy buffer, difference buffer
    eBuffer = new float[int(r/s)];
    dBuffer = new float[15];
    // fEnergy vars: number of averages, energy buffers, difference buffers, onset array
    detectDetail(64);
  }

  void detect(AudioInput in)
  {
    switch ( algorithm )
    {
    case 1: 
      sEnergy(in); 
      break;
    case 2: 
      fEnergy(in); 
      break;
    }
  }

  boolean isOnset()
  {
    return isOnset || isRange(0, feNum-1, feNum/4);
  }

  boolean isOnset(int i)
  {
    return fIsOnset[i];
  }

  boolean isKick()
  {
    return fIsOnset[0];
  }

  boolean isSnare()
  {
    int low = 1;
    int hi = int(feNum/3);
    int thresh = int ( (hi-low)/3 );
    return isRange(low, hi, thresh);
  }

  boolean isHat()
  {
    int low = int(feNum/2);
    int hi = feNum-1;
    int thresh = int( (hi-low)/3 );
    return isRange(low, hi, thresh);
  }

  boolean isRange(int low, int high, int threshold)
  {
    int num = 0;
    for (int i = low; i < high+1; i++) if ( isOnset(i) ) num++;
    return num > threshold;
  }

  void detectMode( String algo )
  {
    if ( algo == "SOUND_ENERGY" ) 
    {
      algorithm = 1;
      fIsOnset = new boolean[feNum];
    }
    else if ( algo == "FREQ_ENERGY" ) 
    {
      algorithm = 2;
      valCnt = varCnt = 0;
      valGraph = new float[width];
      varGraph = new float[width];
    }      
  }

  void detectDetail( int num )
  {
    feNum = num;
    spect.averages(num);
    feBuffer = new float[num][int(sampleRate/bins)];
    fdBuffer = new float[num][15];
    fIsOnset = new boolean[num];
  }

  void drawGraph()
  {
    // draw valGraph
    for (int i = 0; i < valCnt; i++)
    {
      stroke(255);
      line(i, 160 - valGraph[i], i, 160 + valGraph[i]);
    }
    // draw varGraph
    for (int i = 0; i < varCnt-1; i++)
    {
      stroke(255);
      line(i, 200 - varGraph[i], i+1, 200 - varGraph[i+1]);
    }

    strokeWeight(4);
    for (int i = 0; i < fIsOnset.length; i++)
    {
      int c = (i%8 == 0) ? color(255, 0, 0) : color(255);
      stroke(c);
      if ( fIsOnset[i] ) line((i*8), 120, (i*8), 200);
    }
  } 

  private void sEnergy(AudioInput in)
  {
    float instant = spect.getLevel(in)*100;
    // compute the average local energy
    float E = average(eBuffer);
    // compute the variance of the energies in eBuffer
    float V = variance(eBuffer, E);
    // compute C using a linear digression of C with V
    float C = (-0.0025714*V) + 1.5142857;
    float diff = max(instant - C*E, 0);
    pushVal(diff);
    // find the average of only the positive values in dBuffer
    float dAvg = specAverage(dBuffer);
    // filter negative values
    float diff2 = max(diff - dAvg, 0);
    pushVar(diff2);
    if ( isOnset && currFrame == frameCount ) isOnset = true;
    else if ( currFrame < frameCount && isOnset ) isOnset = false;
    else isOnset = diff2 > 0 && instant > 2;
    shift(eBuffer, instant);
    shift(dBuffer, diff);
    currFrame = frameCount;
  }

  private void fEnergy(AudioInput in)
  {
    float instant, E, V, C, diff, dAvg, diff2;
    spect.getSpectrum(in);
    for (int i = 0; i < feNum; i++)
    {
      instant = (i+1)*spect.averages[i]*2000;
      E = average(feBuffer[i]);
      V = variance(feBuffer[i], E);
      C = (-0.0025714*V) + 1.5142857;
      diff = max(instant - C*E, 0);
      dAvg = specAverage(fdBuffer[i]);
      diff2 = max(diff - dAvg, 0);
      if ( fIsOnset[i] && currFrame == frameCount ) fIsOnset[i] = true;
      else if ( currFrame < frameCount && fIsOnset[i] ) fIsOnset[i] = false;
      else fIsOnset[i] = diff2 > 0 && instant > 2;
      shift(feBuffer[i], instant);
      shift(fdBuffer[i], diff);
    }
    currFrame = frameCount;    
  }

  private void pushVal(float v)
  {
    // println(valCnt);
    if ( valCnt == width)
    {
      valCnt = 0;
      valGraph = new float[width];
    }
    valGraph[valCnt] = v;
    valCnt++;
  }

  private void pushVar(float v)
  {
    // println(valCnt);
    if ( varCnt == width)
    {
      varCnt = 0;
      varGraph = new float[width];
    }
    varGraph[varCnt] = v;
    varCnt++;
  } 

  private void shift(float[] arr, float val)
  {
    arraycopy(arr, 0, arr, 1, arr.length-1);
    arr[0] = val;
  }

  private float average(float[] arr)
  {
    float avg = 0;
    for (int i = 0; i < arr.length; i++) avg += arr[i];
    avg /= arr.length;
    return avg;
  }

  private float specAverage(float[] arr)
  {
    float avg = 0;
    float num = 0;
    for (int i = 0; i < arr.length; i++) 
    {
      if ( arr[i] > 0 ) 
      {
        avg += arr[i];
        num++;
      }
    }
    if ( num > 0 ) avg /= num;
    return avg;
  }

  private float variance(float[] arr, float val)
  {
    float V = 0;
    for (int i = 0; i < arr.length; i++) V += pow(arr[i]-val, 2);
    V /= arr.length;
    return V;
  }   
}

//******************************************************************************

void EQ() {
  equalize = (equalize ? false : true);
  analysis.equalizer(equalize);
}

void Limits() {
  if (limit)
  {
    analysis.noLimits();
    limit = false;
  }
  else
  {
    analysis.limits();
    limit = true;
  }
}

void Average() {
  useAvg = (useAvg ? false : true);
}

void Mode(boolean mode) {
  if (mode){
    frame.detectMode("SOUND_ENERGY");
  }
  else{
    frame.detectMode("FREQ_ENERGY");
  }
}

void Damping(float dampAmt) {
  analysis.damp(dampAmt);
}

void Adjust(float adjust) {
  adjust_value = adjust;
}

